@@ -6,5 +6,9 @@ from tf_geometric.nn.conv.gat import *
 from tf_geometric.nn.kernel.map_reduce import *
 from tf_geometric.nn.pool.common_pool import *
 from tf_geometric.nn.pool.topk_pool import *
+from tf_geometric.nn.conv.chebnet import *
+from tf_geometric.nn.conv.sgc import *
+from tf_geometric.nn.conv.tagcn import *
+from tf_geometric.nn.conv.graph_sage import *
 
 
